#include "CyanBusTX.hpp"

// C++ system level
#include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
#include <cinttypes>
// #include <functional>

// ESP32 specific
#include "esp_err.h"
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/uart.h"

// project specific
// #include ""

// qthing stuff
// #include <qthing>



static const char* TAG = "CyanBusTX";


namespace SiliconTorch {

  namespace CyanBusTX {


    CyanBusTX::CyanBusTX(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate, uint8_t uartChannel) : tx(tx), rx(rx), de(de), re(re), uartChannel(uartChannel) {

      uint64_t bitMask = 0;

      bitMask |= 1ULL << tx;
      bitMask |= 1ULL << de;
      bitMask |= 1ULL << re;

      gpio_config_t gpio = {
        .pin_bit_mask = bitMask,
        .mode         = GPIO_MODE_OUTPUT,
        .pull_up_en   = GPIO_PULLUP_DISABLE,
        .pull_down_en = GPIO_PULLDOWN_DISABLE,
        .intr_type    = GPIO_INTR_DISABLE,        
      };

      // configure outputs
      gpio_config(&gpio);
      txrxEN(false);


      gpio.pin_bit_mask = 1ULL << rx;
      gpio.mode = GPIO_MODE_INPUT;

      // configure inputs
      gpio_config(&gpio);


      uart_config_t uartConfig = {
        .baud_rate           = (int)baudRate,
        .data_bits           = UART_DATA_8_BITS,
        .parity              = UART_PARITY_DISABLE,
        .stop_bits           = UART_STOP_BITS_1,
        .flow_ctrl           = UART_HW_FLOWCTRL_DISABLE,
        .rx_flow_ctrl_thresh = 0,
        .use_ref_tick        = false,
        // .source_clk = UART_SCLK_APB,
      };


      uart_port_t _ch = (uart_port_t)uartChannel;

      ESP_ERROR_CHECK(uart_param_config(_ch, &uartConfig));
      ESP_ERROR_CHECK(uart_set_pin(_ch, tx, rx, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE));
      // uart_intr_config(_ch);

      ESP_ERROR_CHECK(uart_driver_install(_ch, 256, MTU * 2, 0, NULL, 0));  // rxBuf = 256; (hopefully greater than FIFO size also in the future)

      ESP_ERROR_CHECK(uart_set_mode(_ch, UART_MODE_UART));

      // WARNING: RX timeout MUST be > 0 !!
      // It determines after how much symbols pause an event is raised (event queue uses TOUT-feature internally)
      // uart_set_rx_timeout(_ch, 5);
      // TODO: We will only send… so just delete this completely??


      // transmit only ATM!
      txEN(true);
      rxEN(false);
    }


    void CyanBusTX::write(std::string& data, bool _break) {
      write((uint8_t*)(data.c_str()), data.length(), _break);
    }

    void CyanBusTX::write(const uint8_t* data, uint32_t length, bool _break) {

      uint32_t bytes = 0;

      if (_break) bytes = uart_write_bytes           ((uart_port_t)uartChannel, (char*)data, length);
      else        bytes = uart_write_bytes_with_break((uart_port_t)uartChannel, (char*)data, length, 16);  // break length = 16 bits


      // TODO: Set level to DEBUG after dev
      ESP_LOGI(TAG, "Send |bytes|[ %d ]", bytes);

      if (bytes != length) ESP_LOGW(TAG, "Send only |bytes|[ %d ] from a |buffer|[ %d ]", bytes, length);
    }


    void CyanBusTX::sendBreak() {
      // TODO: TEST IF THIS REALLY WORKS??
      // aka is it legal to say "hey, here's a NULL-ptr to my data but you should only send 0 bytes so it should be fine"
      // uart_write_bytes_with_break((uart_port_t)uartChannel, NULL, 0, 16);  // break length = 16 bits


      // OK just play safe ATM:
      std::string empty("");
      write(empty, true);
    }


    void CyanBusTX::setBaudRate(uint32_t baudRate) {
      uart_set_baudrate((uart_port_t)uartChannel, baudRate);
    }

    uint32_t CyanBusTX::getBaudRate() {
      uint32_t baudRate = 0;
      uart_get_baudrate((uart_port_t)uartChannel, &baudRate);
      return baudRate;
    }


    void CyanBusTX::txEN(bool state) {
      gpio_set_level((gpio_num_t)de, state);
    }

    void CyanBusTX::rxEN(bool state) {
      gpio_set_level((gpio_num_t)re, state ^ 1);
    }
    
    void CyanBusTX::txrxEN(bool state) {
      txEN(state);
      rxEN(state);
    }

  }
}
